# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: qcsresults.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


import mimiqcircuits.proto.bitvector_pb2 as bitvector__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10qcsresults.proto\x1a\x0f\x62itvector.proto\"+\n\rComplexDouble\x12\x0c\n\x04real\x18\x01 \x01(\x01\x12\x0c\n\x04imag\x18\x02 \x01(\x01\"-\n\rComplexVector\x12\x1c\n\x04\x64\x61ta\x18\x01 \x03(\x0b\x32\x0e.ComplexDouble\"F\n\x0e\x41mplitudeEntry\x12\x17\n\x03key\x18\x01 \x01(\x0b\x32\n.BitVector\x12\x1b\n\x03val\x18\x02 \x01(\x0b\x32\x0e.ComplexDouble\"\x99\x02\n\nQCSResults\x12\x11\n\tsimulator\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t\x12\x12\n\nfidelities\x18\x03 \x03(\x01\x12\x15\n\ravggateerrors\x18\x04 \x03(\x01\x12\x1b\n\x07\x63states\x18\x05 \x03(\x0b\x32\n.BitVector\x12\x1f\n\x07zstates\x18\x06 \x03(\x0b\x32\x0e.ComplexVector\x12#\n\namplitudes\x18\x07 \x03(\x0b\x32\x0f.AmplitudeEntry\x12)\n\x07timings\x18\x08 \x03(\x0b\x32\x18.QCSResults.TimingsEntry\x1a.\n\x0cTimingsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x01:\x02\x38\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'qcsresults_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_QCSRESULTS_TIMINGSENTRY']._options = None
  _globals['_QCSRESULTS_TIMINGSENTRY']._serialized_options = b'8\001'
  _globals['_COMPLEXDOUBLE']._serialized_start=37
  _globals['_COMPLEXDOUBLE']._serialized_end=80
  _globals['_COMPLEXVECTOR']._serialized_start=82
  _globals['_COMPLEXVECTOR']._serialized_end=127
  _globals['_AMPLITUDEENTRY']._serialized_start=129
  _globals['_AMPLITUDEENTRY']._serialized_end=199
  _globals['_QCSRESULTS']._serialized_start=202
  _globals['_QCSRESULTS']._serialized_end=483
  _globals['_QCSRESULTS_TIMINGSENTRY']._serialized_start=437
  _globals['_QCSRESULTS_TIMINGSENTRY']._serialized_end=483
# @@protoc_insertion_point(module_scope)
